# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT

import re
import time

from . import profile

__all__ = []


METHOD_NAME_SCRUBBER = re.compile(r'\W|^(?=\d)')
UNIT_NAME_TO_FUNC_REPLACEMENTS = (
    ('/', ' per '),
    ('%', 'percent'),
    ('*', ' times '))

CRC_START = 0
CRC_TABLE = (
    0x0000, 0xcc01, 0xd801, 0x1400, 0xf001, 0x3c00, 0x2800, 0xe401,
    0xa001, 0x6c00, 0x7800, 0xb401, 0x5000, 0x9c01, 0x8801, 0x4400)


def scrub_method_name(method_name, convert_units=False):
    """Create a valid Python name out of *method_name*"""
    if convert_units:
        for replace_from, replace_to in UNIT_NAME_TO_FUNC_REPLACEMENTS:
            method_name = method_name.replace(replace_from, str(replace_to))

    return METHOD_NAME_SCRUBBER.sub('_', method_name)


def get_mesg_type(mesg_name_or_num):
    """
    Get a :class:`fitdecode.MessageType` from ``profile``, by its name (`str`)
    or its global number (`int`).

    Raise `ValueError` if type was not found.
    """
    # assume mesg_num first
    try:
        return profile.MESSAGE_TYPES[mesg_name_or_num]
    except KeyError:
        pass

    for mesg_type in profile.MESSAGE_TYPES.values():
        if mesg_name_or_num == mesg_type.name:
            return mesg_type

    raise ValueError(f'message type "{mesg_name_or_num}" not found')


def get_mesg_num(mesg_name):
    """
    Get the global number of a message as defined in ``profile``, by its name

    Raise `ValueError` if type was not found.
    """
    for mesg_type in profile.MESSAGE_TYPES.values():
        if mesg_name == mesg_type.name:
            return mesg_type.mesg_num

    raise ValueError(f'message type "{mesg_name}" not found')


def get_mesg_field(mesg_name_or_num, field_name_or_num):
    """
    Get the :class:`fitdecode.types.Field` object of a particular field from a
    particular message.

    Raise `ValueError` if message or field was not found.
    """
    mesg_type = get_mesg_type(mesg_name_or_num)
    for field in mesg_type.fields:
        if field_name_or_num in (field.def_num, field.name):
            return field

    raise ValueError(
        f'field "{field_name_or_num}" not found in '
        f'message "{mesg_name_or_num}"')


def get_mesg_field_num(mesg_name_or_num, field_name):
    """
    Get the definition number of a particular field from a particular message.

    Raise `ValueError` if message or field was not found.
    """
    mesg_type = get_mesg_type(mesg_name_or_num)
    for field in mesg_type.fields:
        if field.name == field_name:
            return field.def_num

    raise ValueError(
        f'field "{field_name}" not found in message "{mesg_name_or_num}"')


def get_field_type(field_name):
    """
    Get :class:`fitdecode.FieldType` by name from ``profile``.

    Raise `ValueError` if type was not found.
    """
    try:
        return profile.FIELD_TYPE[field_name]
    except KeyError:
        raise ValueError(f'field type "{field_name}" not found')


def compute_crc(byteslike, *, crc=CRC_START, start=0, end=None):
    """
    Compute the CRC as per FIT definition, of *byteslike* object, from offset
    *start* (included) to *end* (excluded)
    """
    if not end:
        end = len(byteslike)

    if start >= end:
        assert 0
        return crc

    # According to some performance tests, A is always (at least slightly)
    # faster than B, either with a high number of calls to this fonction, and/or
    # with a high number of "for" iterations (CPython 3.6.5 x64 on Windows).
    #
    # A. for byte in memoryview(byteslike)[start:end]:
    #        # ...
    #
    # B. for idx in range(start, end):
    #        byte = byteslike[idx]
    #        # ...

    for byte in memoryview(byteslike)[start:end]:
        tmp = CRC_TABLE[crc & 0xf]
        crc = (crc >> 4) & 0x0fff
        crc = crc ^ tmp ^ CRC_TABLE[byte & 0xf]

        tmp = CRC_TABLE[crc & 0xf]
        crc = (crc >> 4) & 0x0fff
        crc = crc ^ tmp ^ CRC_TABLE[(byte >> 4) & 0xf]

    return crc


def blocking_read(istream, size=-1, nonblocking_reads_delay=0.06):
    """
    Read from *istream* and do not return until *size* `bytes` have been read
    unless EOF has been reached.

    Return all the data read so far. The length of the returned data may still
    be less than *size* in case EOF has been reached.

    *nonblocking_reads_delay* specifies the number of seconds (float) to wait
    before trying to read from *istream* again in case `BlockingIOError` has
    been raised during previous call.
    """
    assert size is None or (isinstance(size, int) and not isinstance(size, bool))

    if not size:
        return None

    output = []
    len_read = 0

    def _join():
        glue = '' if isinstance(output[0], str) else b''
        return glue.join(output)

    while True:
        try:
            chunk = istream.read(-1 if size < 0 else size - len_read)

            if size < 0:
                return chunk
            elif not chunk:
                if not output:
                    return chunk
                else:
                    return _join()
            else:
                assert size > 0
                output.append(chunk)
                len_read += len(chunk)
                if len_read >= size:
                    return _join()
        except BlockingIOError:
            time.sleep(nonblocking_reads_delay)



# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT

from .__meta__ import (
    __version__, version_info,
    __title__, __fancy_title__, __description__, __url__,
    __license__, __author__, __copyright__)

from .exceptions import *
from .records import *
from .reader import *
from .processors import *

from . import types
from . import profile
from . import utils
from . import processors
from . import reader



# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT

__version__ = '0.11.0'
version_info = (0, 11, 0)

__title__ = 'fitdecode'
__fancy_title__ = 'fitdecode'
__description__ = 'FIT file parser and decoder'
__url__ = 'https://github.com/polyvertex/fitdecode'
__license__ = 'MIT'
__author__ = 'Jean-Charles Lefebvre'
__author_email__ = 'polyvertex@gmail.com'
__copyright__ = 'Copyright 2018-present Jean-Charles Lefebvre'

__keywords__ = ['fit', 'ant', 'file', 'parse', 'parser', 'decode', 'decoder']



#!/usr/bin/env python
# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT

import argparse
import datetime
import json
import re
import sys
import traceback
import types
from collections import OrderedDict

import fitdecode


class RecordJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, types.GeneratorType):
            return list(obj)

        if isinstance(obj, datetime.time):
            return obj.isoformat()

        if isinstance(obj, datetime.datetime):
            return obj.isoformat()

        if isinstance(obj, fitdecode.FitChunk):
            return OrderedDict((
                ('index', obj.index),
                ('offset', obj.offset),
                ('size', len(obj.bytes))))

        if isinstance(obj, fitdecode.types.FieldDefinition):
            return OrderedDict((
                ('name', obj.name),
                ('def_num', obj.def_num),
                ('type_name', obj.type.name),
                ('base_type_name', obj.base_type.name),
                ('size', obj.size)))

        if isinstance(obj, fitdecode.types.DevFieldDefinition):
            return OrderedDict((
                ('name', obj.name),
                ('dev_data_index', obj.dev_data_index),
                ('def_num', obj.def_num),
                ('type_name', obj.type.name),
                ('size', obj.size)))

        if isinstance(obj, fitdecode.types.FieldData):
            return OrderedDict((
                ('name', obj.name),
                ('value', obj.value),
                ('units', obj.units if obj.units else ''),
                ('def_num', obj.def_num),
                ('raw_value', obj.raw_value)))

        if isinstance(obj, fitdecode.FitHeader):
            crc = obj.crc if obj.crc else 0
            return OrderedDict((
                ('frame_type', 'header'),
                ('header_size', obj.header_size),
                ('proto_ver', obj.proto_ver),
                ('profile_ver', obj.profile_ver),
                ('body_size', obj.body_size),
                ('crc', f'{crc:#06x}'),
                ('crc_matched', obj.crc_matched),
                ('chunk', obj.chunk)))

        if isinstance(obj, fitdecode.FitCRC):
            return OrderedDict((
                ('frame_type', 'crc'),
                ('crc', f'{obj.crc:#06x}'),
                ('matched', obj.matched),
                ('chunk', obj.chunk)))

        if isinstance(obj, fitdecode.FitDefinitionMessage):
            return OrderedDict((
                ('frame_type', 'definition_message'),
                ('name', obj.name),
                ('header', OrderedDict((
                    ('local_mesg_num', obj.local_mesg_num),
                    ('time_offset', obj.time_offset),
                    ('is_developer_data', obj.is_developer_data)))),
                ('global_mesg_num', obj.global_mesg_num),
                ('endian', obj.endian),
                ('field_defs', obj.field_defs),
                ('dev_field_defs', obj.dev_field_defs),
                ('chunk', obj.chunk)))

        if isinstance(obj, fitdecode.FitDataMessage):
            return OrderedDict((
                ('frame_type', 'data_message'),
                ('name', obj.name),
                ('header', OrderedDict((
                    ('local_mesg_num', obj.local_mesg_num),
                    ('time_offset', obj.time_offset),
                    ('is_developer_data', obj.is_developer_data)))),
                ('fields', obj.fields),
                ('chunk', obj.chunk)))

        # fall back to original to raise a TypeError
        return super().default(obj)


def parse_filter_args(arg_parser, filter_opt):
    FILTER_DESC = re.compile(r'^\s*([\+\-]?)\s*([^\s]+)\s*$', re.A)

    if not filter_opt:
        return filter_opt, None

    filtr = {}  # {msg_num: bool_include}
    default_include_policy = False

    for desc in filter_opt:
        msg = None
        rem = FILTER_DESC.fullmatch(desc)
        if rem:
            include = False if rem[1] and rem[1] == '-' else True
            msg = rem[2].lower()

            if not include:
                default_include_policy = True

            try:
                msg = fitdecode.utils.get_mesg_num(msg)
            except ValueError:
                try:
                    msg = int(msg, base=0)
                except ValueError:
                    msg = None

        if msg is None:
            arg_parser.error(f'malformed filter: "{desc}"')
            sys.exit(1)

        filtr[msg] = include

    return filtr, default_include_policy


def parse_args(args=None):
    parser = argparse.ArgumentParser(
        description='Dump a FIT file to JSON format',
        epilog=f'fitdecode version {fitdecode.__version__}',
        allow_abbrev=False)

    parser.add_argument(
        '--output', '-o', type=argparse.FileType(mode='wt', encoding='utf-8'),
        default='-',
        help='File to output data into (defaults to stdout)')

    parser.add_argument(
        '--pretty', action='store_true',
        help='Prettify JSON output.')

    parser.add_argument(
        '--nocrc', action='store_const',
        const=fitdecode.CrcCheck.DISABLED,
        default=fitdecode.CrcCheck.WARN,
        help='Some devices seem to write invalid CRC\'s, ignore these.')

    parser.add_argument(
        '--nodef', action='store_true',
        help='Do not output FIT local message definitions.')

    parser.add_argument(
        '--nounk', action='store_true',
        help='Do not output unknown FIT messages (e.g. "unknown_140")')

    parser.add_argument(
        '--filter', '-f', action='append',
        help=(
            'Message name(s) (or global numbers) to filter-in or out, '
            'depending on sign prefix.  Examples: "-record" to exclude record '
            'messages; "+file_id" or "file_id" to include file_id messages.'))

    parser.add_argument(
        'infile', metavar='FITFILE', type=argparse.FileType(mode='rb'),
        help='Input .FIT file (use - for stdin)')

    options = parser.parse_args(args)
    options.filter, options.default_filter = \
        parse_filter_args(parser, options.filter)

    return options


def main(args=None):
    options = parse_args(args)

    frames = []

    try:
        with fitdecode.FitReader(
                options.infile,
                processor=fitdecode.StandardUnitsDataProcessor(),
                check_crc=options.nocrc,
                keep_raw_chunks=True) as fit:
            for frame in fit:
                if (options.nodef and
                        frame.frame_type == fitdecode.FIT_FRAME_DEFINITION):
                    continue

                if (options.nounk and
                        frame.frame_type in (
                            fitdecode.FIT_FRAME_DEFINITION,
                            fitdecode.FIT_FRAME_DATA) and
                        frame.mesg_type is None):
                    continue

                if (options.filter and
                        frame.frame_type in (
                            fitdecode.FIT_FRAME_DEFINITION,
                            fitdecode.FIT_FRAME_DATA)):
                    try:
                        include = options.filter[frame.global_mesg_num]
                    except KeyError:
                        include = options.default_filter

                    if not include:
                        continue

                frames.append(frame)
    except Exception:
        print(
            'WARNING: the following error occurred while parsing FIT file. '
            'Output file might be incomplete or corrupted.',
            file=sys.stderr)
        print('', file=sys.stderr)
        traceback.print_exc()

    indent = '\t' if options.pretty else None
    json.dump(frames, fp=options.output, cls=RecordJSONEncoder, indent=indent)

    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT

import datetime

from . import profile
from .utils import scrub_method_name

__all__ = [
    'FIT_UTC_REFERENCE', 'FIT_DATETIME_MIN',
    'DataProcessorBase', 'DefaultDataProcessor', 'StandardUnitsDataProcessor']


#: Datetimes (uint32) represent seconds since this ``FIT_UTC_REFERENCE``
#: (unix timestamp for UTC 00:00 Dec 31 1989).
FIT_UTC_REFERENCE = 631065600

#: ``date_time`` typed fields for which value is below ``FIT_DATETIME_MIN``
#: represent the number of seconds elapsed since device power on.
FIT_DATETIME_MIN = 0x10000000


class DataProcessorBase:
    """
    Data processing base class.

    This class does nothing. It is meant to be derived.

    The following methods are called by :class:`fitdecode.FitReader`:

    * `on_header`, each time a :class:`fitdecode.FitHeader` is reached
    * `on_crc`, each time a :class:`fitdecode.FitCRC` (the FIT footer) is
      reached

    This is convenient if you wish to reset some context-sensitive state in-
    between two chained FIT files for example.

    Bear in mind that a malformed/corrupted file may miss either of these
    entities (header and/or CRC footer).

    Also, the following methods are called (still by
    :class:`fitdecode.FitReader`) for each field of every data message, in that
    order:

    * `on_process_type`
    * `on_process_field`
    * `on_process_unit`
    * `on_process_message`

    By default, the above processor methods call the following methods if they
    exist (hence the aforementioned caching)::

        def process_type_<type_name>(reader, field_data)
        def process_field_<field_name>(reader, field_data)  # could be unknown_XYZ but NOT recommended  # noqa
        def process_units_<unit_name>(reader, field_data)
        def process_message_<mesg_name>(reader, data_message)

    ``process_*`` methods are not expected to return any value and may alter
    the content of the passed *field_data* (:class:`fitdecode.FieldData`) and
    *data_message* (:class:`fitdecode.FitDataMessage`) arguments if needed.

    .. seealso:: `DefaultDataProcessor`, `StandardUnitsDataProcessor`
    """

    def __init__(self):
        self._method_cache = {}

    def on_header(self, reader, fit_header):
        pass

    def on_crc(self, reader, fit_crc):
        pass

    def on_process_type(self, reader, field_data):
        self._run_processor(
            f'process_type_{field_data.type.name}',
            reader, field_data)

    def on_process_field(self, reader, field_data):
        if field_data.name:
            self._run_processor(
                f'process_field_{field_data.name}',
                reader, field_data)

    def on_process_unit(self, reader, field_data):
        if field_data.units:
            self._run_processor(
                f'process_units_{field_data.units}',
                reader, field_data)

    def on_process_message(self, reader, data_message):
        self._run_processor(
            f'process_message_{data_message.def_mesg.name}',
            reader, data_message)

    def _run_processor(self, method_name, reader, data):
        method = self._resolve_method(method_name)
        if method is not None:
            method(reader, data)

    def _resolve_method(self, method_name):
        try:
            return self._method_cache[method_name]
        except KeyError:
            pass

        scrubbed_method_name = scrub_method_name(method_name)
        method = getattr(self, scrubbed_method_name, None)

        self._method_cache[method_name] = method

        return method


class DefaultDataProcessor(DataProcessorBase):
    """
    This is the default data processor used by :class:`fitdecode.FitReader`. It
    derives from :class:`DataProcessorBase`.

    This data processor converts some raw values to more comfortable ones.

    .. seealso:: `StandardUnitsDataProcessor`, `DataProcessorBase`
    """

    def __init__(self):
        super().__init__()

    def process_type_bool(self, reader, field_data):
        """Just `bool` any ``bool`` typed FIT field unless value is `None`"""
        if field_data.value is not None:
            field_data.value = bool(field_data.value)

    def process_type_date_time(self, reader, field_data):
        """
        Convert ``date_time`` typed field values into `datetime.datetime` object
        if possible.

        That is, if value is not `None` and greater or equal than
        `FIT_DATETIME_MIN`.

        The resulting `datetime.datetime` object is timezone-aware (UTC).
        """
        if (field_data.value is not None and
                field_data.value >= FIT_DATETIME_MIN):
            field_data.value = datetime.datetime.fromtimestamp(
                FIT_UTC_REFERENCE + field_data.value,
                datetime.timezone.utc)
            field_data.units = None  # units were 's', set to None

    def process_type_local_date_time(self, reader, field_data):
        """
        Convert ``date_time`` typed field values into `datetime.datetime` object
        unless value is `None`.

        The resulting `datetime.datetime` object **IS NOT** timezone-aware, but
        this method assumes UTC at object construction to ensure consistency.
        """
        if field_data.value is not None:
            # This value was created on the device using its local timezone.
            # Unless we know that timezone, this value won't be correct.
            # However, if we assume UTC, at least it'll be consistent.
            field_data.value = datetime.datetime.fromtimestamp(
                FIT_UTC_REFERENCE + field_data.value,
                datetime.timezone.utc)
            field_data.units = None

    def process_type_localtime_into_day(self, reader, field_data):
        """
        Convert ``localtime_into_day`` typed field values into `datetime.time`
        object unless value is `None`.
        """
        if field_data.value is not None:
            # Values larger or equal to 86400 should not be possible.
            # Additionally, if the value is exactly 86400, it will lead to an
            # error when trying to create the time with datetime.time(24, 0, 0).
            #
            # E.g. Garmin does add "sleep_time": 86400 to its fit files, which
            # causes an error if not properly handled.
            if field_data.value >= 86400:
                field_data.value = datetime.time.max
            else:
                m, s = divmod(field_data.value, 60)
                h, m = divmod(m, 60)
                field_data.value = datetime.time(h, m, s)
            field_data.units = None

    def process_message_hr(self, reader, data_message):
        """
        Convert populated ``event_timestamp`` component values of the ``hr`` to
        `datetime.datetime` objects
        """
        # We want to convert only populated *event_timestamp* fields that were
        # originally computed from the *event_timestamp_12* value
        if data_message.has_field(profile.FIELD_NUM_HR_EVENT_TIMESTAMP_12):
            for field_data in data_message.get_fields(
                    profile.FIELD_NUM_HR_EVENT_TIMESTAMP):
                if field_data is not None:
                    field_data.value = datetime.datetime.fromtimestamp(
                        FIT_UTC_REFERENCE + field_data.value,
                        datetime.timezone.utc)
                    field_data.units = None  # units were 's', set to None


class StandardUnitsDataProcessor(DefaultDataProcessor):
    """
    A `DefaultDataProcessor` that also:

    * Converts ``distance`` and ``total_distance`` fields to ``km``
      (standard's default is ``m``)
    * Converts all ``speed`` and ``*_speeds`` fields (by name) to ``km/h``
      (standard's default is ``m/s``)
    * Converts GPS coordinates (i.e. FIT's semicircles type) to ``deg``

    .. seealso:: `DefaultDataProcessor`, `DataProcessorBase`
    """

    def __init__(self):
        super().__init__()

    def on_process_field(self, reader, field_data):
        """
        Convert all ``*_speed`` fields using `process_field_speed`.

        All other units will use the default method.
        """
        if field_data.name and field_data.name.endswith('_speed'):
            self.process_field_speed(reader, field_data)
        else:
            super().on_process_field(reader, field_data)

    def process_field_distance(self, reader, field_data):
        if field_data.value is not None:
            field_data.value /= 1000.0
        field_data.units = 'km'

    def process_field_total_distance(self, reader, field_data):
        self.process_field_distance(reader, field_data)

    def process_field_speed(self, reader, field_data):
        if field_data.value is not None:
            factor = 60.0 * 60.0 / 1000.0

            # record.enhanced_speed field can be a tuple...
            # see https://github.com/dtcooper/python-fitparse/issues/62
            if isinstance(field_data.value, (tuple, list)):
                field_data.value = tuple(x * factor for x in field_data.value)
            else:
                field_data.value *= factor

        field_data.units = 'km/h'

    def process_units_semicircles(self, reader, field_data):
        if field_data.value is not None:
            field_data.value *= 180.0 / (2 ** 31)
        field_data.units = 'deg'



#!/usr/bin/env python
# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT

import argparse
import datetime
import decimal
import os.path
import re
import sys
import traceback
from collections import OrderedDict

import fitdecode

echo = None


def txt_encode(obj):
    """
    Convert a given *obj* to either a *str* or *PrintableObject* depending on
    which is the most suitable
    """

    if isinstance(obj, PrintableObject):
        return obj

    if isinstance(obj, str):
        return '[' + obj + ']'

    if obj is None:
        return '<none>'

    if isinstance(obj, bool):
        return 'yes' if obj else 'no'

    if isinstance(obj, (int, float, decimal.Decimal)):
        return str(obj)

    if isinstance(obj, datetime.time):
        return obj.isoformat()

    if isinstance(obj, datetime.datetime):
        return obj.isoformat()

    if isinstance(obj, fitdecode.FitChunk):
        return PrintableObject(
            index=obj.index,
            offset=obj.offset,
            size=len(obj.bytes))

    if isinstance(obj, fitdecode.types.FieldDefinition):
        return PrintableObject(
            name=obj.name,
            def_num=obj.def_num,
            type_name=obj.type.name,
            base_type_name=obj.base_type.name,
            size=obj.size)

    if isinstance(obj, fitdecode.types.DevFieldDefinition):
        return PrintableObject(
            name=obj.name,
            dev_data_index=obj.dev_data_index,
            def_num=obj.def_num,
            type_name=obj.type.name,
            size=obj.size)

    if isinstance(obj, fitdecode.types.FieldData):
        return PrintableObject(
            name=obj.name,
            value=obj.value,
            units=obj.units if obj.units else '',
            def_num=obj.def_num,
            raw_value=obj.raw_value)

    if isinstance(obj, fitdecode.FitHeader):
        crc = obj.crc if obj.crc else 0
        return PrintableObject(
            _label=f'chunk#{obj.chunk.index} - fit_header',
            header_size=obj.header_size,
            proto_ver='(' + ', '.join([str(v) for v in obj.proto_ver]) + ')',
            profile_ver='(' + ', '.join([str(v) for v in obj.profile_ver]) + ')',
            body_size=obj.body_size,
            crc=f'{crc:#06x}',
            crc_matched=obj.crc_matched,
            chunk=obj.chunk)

    if isinstance(obj, fitdecode.FitCRC):
        return PrintableObject(
            _label=f'chunk#{obj.chunk.index} - fit_crc',
            crc=f'{obj.crc:#06x}',
            matched=obj.matched,
            chunk=obj.chunk)

    if isinstance(obj, fitdecode.FitDefinitionMessage):
        return PrintableObject(
            _label=(
                f'chunk#{obj.chunk.index} - fit_definition - {obj.name} '
                f'(loc#{obj.local_mesg_num} glob#{obj.global_mesg_num})'),
            chunk=obj.chunk,
            header=PrintableObject(
                local_mesg_num=obj.local_mesg_num,
                time_offset=obj.time_offset,
                is_developer_data=obj.is_developer_data),
            global_mesg_num=obj.global_mesg_num,
            endian=obj.endian,
            field_defs=obj.field_defs,
            dev_field_defs=obj.dev_field_defs)

    if isinstance(obj, fitdecode.FitDataMessage):
        return PrintableObject(
            _label=(
                f'chunk#{obj.chunk.index} - fit_data - {obj.name} '
                f'(loc#{obj.local_mesg_num} glob#{obj.global_mesg_num})'),
            chunk=obj.chunk,
            header=PrintableObject(
                local_mesg_num=obj.local_mesg_num,
                time_offset=obj.time_offset,
                is_developer_data=obj.is_developer_data),
            global_mesg_num=obj.global_mesg_num,
            fields=obj.fields)

    if __debug__:
        print(type(obj))
        assert 0

    return repr(obj)


class PrintableObject:
    __slots__ = (
        '_label', '_dict', '_max_key_length',
        '_pad', '_key_prefix', '_key_suffix')

    def __init__(self, *, _label=None, _pad=' ', _key_prefix='',
                 _key_suffix='  ', **props):
        self._label = _label
        self._dict = OrderedDict()
        self._max_key_length = 0
        self._pad = _pad
        self._key_prefix = _key_prefix
        self._key_suffix = _key_suffix

        for key, value in props.items():
            # to avoid potential collision with PrintableObject members
            # (see __setattr__)
            assert key[0] != '_'

            self._dict[key] = value
            if len(key) > self._max_key_length:
                self._max_key_length = len(key)

    def __iter__(self):
        for key, value in self._dict.items():
            name = self._key_prefix
            name += key.ljust(self._max_key_length, self._pad)
            name += self._key_suffix

            yield name, value

    def __getattr__(self, name):
        try:
            return self._dict[name]
        except KeyError:
            raise AttributeError

    def __setattr__(self, name, value):
        if name[0] == '_':
            super().__setattr__(name, value)
        elif name not in self._dict:
            raise AttributeError
        else:
            self._dict[name] = value


def global_stats(frames, options):
    if options.filter:
        filter_str = []
        for msg_num, include in options.filter.items():
            include = '-' if not include else '+'
            msg_name = fitdecode.utils.get_mesg_type(msg_num).name
            filter_str.append(f'{include}{msg_name}#{msg_num}')
        filter_str = '[' + ', '.join(filter_str) + ']'
    else:
        filter_str = '[]'

    stats = PrintableObject(
        _label='TXT',
        name=os.path.basename(options.infile.name),
        filter=filter_str,
        frames=len(frames),
        size=0,
        missing_headers=0,
        fit_files=[])

    stats_got_header = False
    for frame in frames:
        stats.size += len(frame.chunk.bytes)

        if isinstance(frame, fitdecode.FitHeader):
            stats_got_header = True
            stats.fit_files.append(PrintableObject(
                definition_messages=0,
                data_messages=0,
                has_footer=False,
                header_crc_matched=frame.crc_matched,
                footer_crc_matched=False))
            continue

        if not stats_got_header:
            stats.missing_headers += 1
            continue

        curr_file = stats.fit_files[-1]

        if isinstance(frame, fitdecode.FitCRC):
            stats_got_header = False
            curr_file.has_footer = True
            curr_file.footer_crc_matched = frame.matched

        elif isinstance(frame, fitdecode.FitDefinitionMessage):
            curr_file.definition_messages += 1

        elif isinstance(frame, fitdecode.FitDataMessage):
            curr_file.data_messages += 1

    if options.strip:
        stats.fit_files = len(stats.fit_files)

    return stats


def txt_print(obj, *, indent='\t', level=0):

    def _recurse(subobj, sublevel=level):
        txt_print(subobj, indent=indent, level=sublevel)

    def _p(*values, end='\n'):
        echo(indent * level, *values, sep='', end=end)

    if isinstance(obj, str):
        _p(obj)

    elif isinstance(obj, (tuple, list)):
        # first pass to check if we can keep this list on a single line
        one_line = True
        for value in obj:
            if value is not None and not isinstance(
                    value, (bool, int, float, decimal.Decimal)):
                one_line = False
                break

        if one_line:
            _p('[' + ', '.join([txt_encode(v) for v in obj]) + ']')
        else:
            first = True
            for value in obj:
                if first:
                    first = False
                else:
                    _p('-')
                _recurse(value)

    elif isinstance(obj, PrintableObject):
        if obj._label:
            _p(obj._label)
            obj._label = None
            _recurse(obj, sublevel=level + 1)
        else:
            for key, value in obj:
                if isinstance(value, str):
                    _p(key, value)
                elif isinstance(value, (tuple, list)):
                    _p(f'{key.rstrip()} ({len(value)})')
                    _recurse(value, sublevel=level + 1)
                else:
                    value = txt_encode(value)
                    if isinstance(value, str):
                        _p(key, value)
                    else:
                        _p(key.rstrip())
                        _recurse(value, sublevel=level + 1)

    else:
        _recurse(txt_encode(obj))


def parse_filter_args(arg_parser, filter_opt):
    FILTER_DESC = re.compile(r'^\s*([\+\-]?)\s*([^\s]+)\s*$', re.A)

    if not filter_opt:
        return filter_opt, None

    filtr = {}  # {msg_num: bool_include}
    default_include_policy = False

    for desc in filter_opt:
        msg = None
        rem = FILTER_DESC.fullmatch(desc)
        if rem:
            include = False if rem[1] and rem[1] == '-' else True
            msg = rem[2].lower()

            if not include:
                default_include_policy = True

            try:
                msg = fitdecode.utils.get_mesg_num(msg)
            except ValueError:
                try:
                    msg = int(msg, base=0)
                except ValueError:
                    msg = None

        if msg is None:
            arg_parser.error(f'malformed filter: "{desc}"')
            sys.exit(1)

        filtr[msg] = include

    return filtr, default_include_policy


def parse_args(args=None):
    parser = argparse.ArgumentParser(
        description='Dump a FIT file to TXT format that ease debugging',
        epilog=f'fitdecode version {fitdecode.__version__}',
        allow_abbrev=False)

    parser.add_argument(
        '--output', '-o', type=argparse.FileType(mode='wt', encoding='utf-8'),
        default='-',
        help='File to output data into (defaults to stdout)')

    parser.add_argument(
        '--nocrc', action='store_const',
        const=fitdecode.CrcCheck.DISABLED,
        default=fitdecode.CrcCheck.WARN,
        help='Some devices seem to write invalid CRC\'s, ignore these')

    parser.add_argument(
        '--nodef', action='store_true',
        help='Do not output FIT local message definitions')

    parser.add_argument(
        '--nounk', action='store_true',
        help='Do not output unknown FIT messages (e.g. "unknown_140")')

    parser.add_argument(
        '--strip', action='store_true',
        help='Do not output the extended global stats in header')

    parser.add_argument(
        '--filter', '-f', action='append',
        help=(
            'Message name(s) (or global numbers) to filter-in or out, '
            'depending on sign prefix.  Examples: "-record" to exclude record '
            'messages; "+file_id" or "file_id" to include file_id messages.'))

    parser.add_argument(
        'infile', metavar='FITFILE', type=argparse.FileType(mode='rb'),
        help='Input .FIT file (use - for stdin)')

    options = parser.parse_args(args)
    options.filter, options.default_filter = \
        parse_filter_args(parser, options.filter)

    return options


def main(args=None):
    options = parse_args(args)

    def _echo(*objects, sep=' ', end='\n', file=options.output, flush=False):
        print(*objects, sep=sep, end=end, file=file, flush=flush)

    def _echo_separator():
        _echo('')
        _echo('*' * 80)
        _echo('')
        _echo('')

    global echo
    echo = _echo

    # fully parse input file and filter out the unwanted messages
    frames = []
    exception_msg = None
    try:
        with fitdecode.FitReader(
                options.infile,
                processor=fitdecode.StandardUnitsDataProcessor(),
                check_crc=options.nocrc,
                keep_raw_chunks=True) as fit:
            for frame in fit:
                if (options.nodef and
                        frame.frame_type == fitdecode.FIT_FRAME_DEFINITION):
                    continue

                if (options.nounk and
                        frame.frame_type in (
                            fitdecode.FIT_FRAME_DEFINITION,
                            fitdecode.FIT_FRAME_DATA) and
                        frame.mesg_type is None):
                    continue

                if (options.filter and
                        frame.frame_type in (
                            fitdecode.FIT_FRAME_DEFINITION,
                            fitdecode.FIT_FRAME_DATA)):
                    try:
                        include = options.filter[frame.global_mesg_num]
                    except KeyError:
                        include = options.default_filter

                    if not include:
                        continue

                frames.append(frame)
    except Exception:
        print(
            'WARNING: error(s) occurred while parsing FIT file. '
            'See output file for more info.',
            file=sys.stderr)
        exception_msg = traceback.format_exc()

    # print some statistics as a header
    if not exception_msg:
        txt_print(global_stats(frames, options))
        echo('')
    else:
        echo('ERROR OCCURRED WHILE PARSING', options.infile.name)
        echo('')
        echo(exception_msg)
        echo('')

    # pretty-print the file
    had_frames = False
    for frame in frames:
        if had_frames and isinstance(frame, fitdecode.FitHeader):
            _echo_separator()
        had_frames = True
        txt_print(frame)
        echo('')

    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT

import math
import struct

__all__ = []


class BaseType:
    __slots__ = ('name', 'identifier', 'fmt', 'size', 'parse')

    enum = None  # in case we're treated as a FieldType

    def __init__(self, name, identifier, fmt, parse):
        self.name = name
        self.identifier = identifier
        self.fmt = fmt
        self.size = struct.calcsize(fmt)
        self.parse = parse

    @property
    def type_num(self):
        """"Base Type Number" as per SDK definition"""
        return self.identifier & 0x1F


class FieldType:
    __slots__ = ('name', 'base_type', 'enum')

    def __init__(self, name, base_type, enum=None):
        self.name = name
        self.base_type = base_type
        self.enum = enum


class _FieldAndSubFieldBase:
    __slots__ = ()

    @property
    def base_type(self):
        return self.type if self.is_base_type else self.type.base_type

    @property
    def is_base_type(self):
        return isinstance(self.type, BaseType)

    def render(self, raw_value):
        if self.type.enum and (raw_value in self.type.enum):
            return self.type.enum[raw_value]
        return raw_value


class Field(_FieldAndSubFieldBase):
    __slots__ = (
        'name', 'type', 'def_num', 'scale', 'offset', 'units', 'components',
        'subfields')

    field_type = 'field'

    def __init__(self, name, type, def_num, scale=None, offset=None, units=None,
                 components=None, subfields=None):
        super().__init__()
        self.name = name
        self.type = type  #: `FieldType`
        self.def_num = def_num
        self.scale = scale
        self.offset = offset
        self.units = units
        self.components = components
        self.subfields = subfields


class SubField(_FieldAndSubFieldBase):
    __slots__ = ('name', 'def_num', 'type', 'scale', 'offset', 'units',
                 'components', 'ref_fields')

    field_type = 'subfield'

    def __init__(self, name, def_num, type, scale=None, offset=None, units=None,
                 components=None, ref_fields=None):
        super().__init__()
        self.name = name
        self.def_num = def_num
        self.type = type
        self.scale = scale
        self.offset = offset
        self.units = units
        self.components = components
        self.ref_fields = ref_fields


class DevField(_FieldAndSubFieldBase):
    __slots__ = (
        'dev_data_index', 'name', 'def_num', 'type', 'units',
        'native_field_num',

        # "Inherited" from FitField only to maintain interface compatibility.
        # They are always None.
        'scale', 'offset', 'components', 'subfields')

    field_type = 'devfield'

    def __init__(self, dev_data_index, name, def_num, type, units,
                 native_field_num):
        super().__init__()

        self.dev_data_index = dev_data_index
        self.name = name
        self.def_num = def_num
        self.type = type
        self.units = units
        self.native_field_num = native_field_num

        self.scale = None
        self.offset = None
        self.components = None
        self.subfields = None


class ReferenceField:
    __slots__ = ('name', 'def_num', 'value', 'raw_value')

    def __init__(self, name, def_num, value, raw_value):
        self.name = name
        self.def_num = def_num
        self.value = value
        self.raw_value = raw_value


class ComponentField:
    __slots__ = (
        'name', 'def_num', 'scale', 'offset', 'units', 'accumulate', 'bits',
        'bit_offset')

    field_type = 'component'

    def __init__(self, name, def_num, scale=None, offset=None, units=None,
                 accumulate=None, bits=None, bit_offset=None):
        self.name = name
        self.def_num = def_num
        self.scale = scale
        self.offset = offset
        self.units = units
        self.accumulate = accumulate
        self.bits = bits
        self.bit_offset = bit_offset

    def render(self, raw_value):
        if raw_value is None:
            return None

        # if it's a tuple, then it's a byte array and unpack it as such
        # (only type that uses this is compressed speed/distance)
        if isinstance(raw_value, tuple):
            # Profile.xlsx sometimes contains more components than the read raw
            # value is able to hold (typically the *event_timestamp_12* field in
            # *hr* messages).
            # This test allows to ensure *unpacked_num* is not right-shifted
            # more than possible.
            if self.bit_offset and self.bit_offset >= len(raw_value) << 3:
                raise ValueError()

            unpacked_num = 0

            # unpack byte array as little endian
            for value in reversed(raw_value):
                unpacked_num = (unpacked_num << 8) + value

            raw_value = unpacked_num

        # mask and shift like a normal number
        if isinstance(raw_value, int):
            raw_value = (raw_value >> self.bit_offset) & ((1 << self.bits) - 1)

        return raw_value


class MessageType:
    __slots__ = ('name', 'mesg_num', 'fields')

    def __init__(self, name, mesg_num, fields):
        self.name = name
        self.mesg_num = mesg_num
        self.fields = fields


class FieldDefinition:
    __slots__ = ('field', 'def_num', 'base_type', 'size')

    def __init__(self, field, def_num, base_type, size):
        self.field = field  #: `Field`
        self.def_num = def_num
        self.base_type = base_type
        self.size = size

    @property
    def is_dev(self):
        return False

    @property
    def name(self):
        return self.field.name if self.field else 'unknown_' + str(self.def_num)

    @property
    def type(self):
        return self.field.type if self.field else self.base_type


class DevFieldDefinition:
    __slots__ = ('field', 'dev_data_index', 'base_type', 'def_num', 'size')

    def __init__(self, field, dev_data_index, def_num, size):
        self.field = field
        self.dev_data_index = dev_data_index
        self.def_num = def_num
        self.size = size

        # for dev fields, the base_type and type are always the same
        self.base_type = self.type

    @property
    def is_dev(self):
        return True

    @property
    def name(self):
        if self.field:
            return self.field.name
        else:
            return f'unknown_dev_{self.dev_data_index}_{self.def_num}'

    @property
    def type(self):
        return self.field.type


class FieldData:
    __slots__ = (
        'field_def', 'field', 'parent_field', 'value', 'raw_value', 'units')

    def __init__(self, field_def, field, parent_field, value, raw_value,
                 units=None):
        self.field_def = field_def  #: `FieldDefinition` object
        self.field = field
        self.parent_field = parent_field
        self.value = value
        self.raw_value = raw_value
        self.units = units

        if not self.units and self.field:
            # Default to units on field, otherwise None.
            # NOTE: Not a property since you may want to override this in a data
            # processor
            self.units = self.field.units

    @property
    def name(self):
        """
        Field's name as defined in FIT global profile.

        If name was not found in global profile, a string is created with the
        form: ``unknown_{def_num}`` where ``def_num`` is the field's definition
        number.

        This value is **NOT** compatible with `is_named`.

        .. seealso:: `name_or_num`
        """
        return self.field.name if self.field else 'unknown_%d' % self.def_num

    @property
    def name_or_num(self):
        """
        Field's name as defined in FIT global profile.

        If name was not found in global profile, ``self.def_num`` is returned
        (`int`).

        This value is compatible with `is_named`.

        .. seealso:: `name`
        """
        return self.field.name if self.field else self.def_num

    @property
    def def_num(self):
        """Field's definition number (`int`)"""
        # prefer to return the def_num on the field since field_def may be None
        # if this field is dynamic
        return self.field.def_num if self.field else self.field_def.def_num

    @property
    def base_type(self):
        """Field's `BaseType`"""
        # try field_def's base type, if it doesn't exist, this is a dynamically
        # added field, so field doesn't be None
        if self.field_def:
            return self.field_def.base_type
        else:
            return self.field.base_type

    @property
    def is_base_type(self):
        """Field's `BaseType`"""
        return self.field.is_base_type if self.field else True

    @property
    def type(self):
        return self.field.type if self.field else self.base_type

    @property
    def field_type(self):
        return self.field.field_type if self.field else 'field'

    @property
    def is_expanded(self):
        """
        Flag to indicate whether this field has been generated through expansion
        """
        return not self.field_def

    def is_named(self, name_or_num):
        """
        Check if this field has the specified name (`str`) or definition number
        (`int`)
        """
        if self.field:
            if name_or_num in (self.field.def_num, self.field.name):
                return True

        if self.parent_field:
            if name_or_num in (self.parent_field.def_num, self.parent_field.name):
                return True

        if self.field_def:
            if name_or_num == self.field_def.def_num:
                return True

        return False


def parse_string(byteslike):
    try:
        s = byteslike[:byteslike.index(0x00)]
    except ValueError:
        # FIT specification defines the 'string' type as follows: "Null
        # terminated string encoded in UTF-8 format".
        #
        # However 'string' values are not always null-terminated when encoded,
        # according to FIT files created by Garmin devices (e.g. DEVICE.FIT file
        # from a fenix3).
        #
        # So in order to be more flexible, in case index() could not find any
        # null byte, we just decode the whole bytes-like object.
        s = byteslike

    return s.decode(encoding='utf-8', errors='replace') or None


BASE_TYPE_BYTE = BaseType(
    name='byte', identifier=0x0d, fmt='B',
    parse=lambda x: None if all(b == 0xff for b in x) else x)


BASE_TYPES = {
    0x00: BaseType(name='enum', identifier=0x00, fmt='B', parse=lambda x: None if x == 0xff else x),  # noqa: E501
    0x01: BaseType(name='sint8', identifier=0x01, fmt='b', parse=lambda x: None if x == 0x7f else x),  # noqa: E501
    0x02: BaseType(name='uint8', identifier=0x02, fmt='B', parse=lambda x: None if x == 0xff else x),  # noqa: E501
    0x83: BaseType(name='sint16', identifier=0x83, fmt='h', parse=lambda x: None if x == 0x7fff else x),  # noqa: E501
    0x84: BaseType(name='uint16', identifier=0x84, fmt='H', parse=lambda x: None if x == 0xffff else x),  # noqa: E501
    0x85: BaseType(name='sint32', identifier=0x85, fmt='i', parse=lambda x: None if x == 0x7fffffff else x),  # noqa: E501
    0x86: BaseType(name='uint32', identifier=0x86, fmt='I', parse=lambda x: None if x == 0xffffffff else x),  # noqa: E501
    0x07: BaseType(name='string', identifier=0x07, fmt='s', parse=parse_string),
    0x88: BaseType(name='float32', identifier=0x88, fmt='f', parse=lambda x: None if math.isnan(x) else x),  # noqa: E501
    0x89: BaseType(name='float64', identifier=0x89, fmt='d', parse=lambda x: None if math.isnan(x) else x),  # noqa: E501
    0x0a: BaseType(name='uint8z', identifier=0x0a, fmt='B', parse=lambda x: None if x == 0 else x),  # noqa: E501
    0x8b: BaseType(name='uint16z', identifier=0x8b, fmt='H', parse=lambda x: None if x == 0 else x),  # noqa: E501
    0x8c: BaseType(name='uint32z', identifier=0x8c, fmt='I', parse=lambda x: None if x == 0 else x),  # noqa: E501
    0x0d: BASE_TYPE_BYTE,
    0x8e: BaseType(name='sint64', identifier=0x8e, fmt='q', parse=lambda x: None if x == 0x7fffffffffffffff else x),  # noqa: E501
    0x8f: BaseType(name='uint64', identifier=0x8f, fmt='Q', parse=lambda x: None if x == 0xffffffffffffffff else x),  # noqa: E501
    0x90: BaseType(name='uint64z', identifier=0x90, fmt='Q', parse=lambda x: None if x == 0 else x)}  # noqa: E501



# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT

import itertools

__all__ = [
    'FitChunk', 'FitHeader', 'FitCRC', 'FitDefinitionMessage', 'FitDataMessage',
    'FIT_FRAME_HEADER', 'FIT_FRAME_CRC',
    'FIT_FRAME_DEFINITION', 'FIT_FRAME_DATA',
    'FIT_FRAME_DEFMESG', 'FIT_FRAME_DATAMESG']


_UNSET = object()

FIT_FRAME_HEADER = 1
FIT_FRAME_CRC = 2
FIT_FRAME_DEFINITION = 3
FIT_FRAME_DATA = 4

# keep aliases for backward compatibility
FIT_FRAME_DEFMESG = FIT_FRAME_DEFINITION
FIT_FRAME_DATAMESG = FIT_FRAME_DATA


class FitChunk:
    __slots__ = ('index', 'offset', 'bytes')

    def __init__(self, index, offset, bytes):
        self.index = index    #: zero-based index of this frame in the file
        self.offset = offset  #: the offset at which this frame starts in the file
        self.bytes = bytes    #: the frame itself as a `bytes` object


class FitHeader:
    frame_type = FIT_FRAME_HEADER

    __slots__ = (
        'header_size', 'proto_ver', 'profile_ver', 'body_size',
        'crc', 'crc_matched', 'chunk')

    def __init__(
            self, header_size, proto_ver, profile_ver, body_size, crc,
            crc_matched, chunk):
        self.header_size = header_size
        self.proto_ver = proto_ver
        self.profile_ver = profile_ver
        self.body_size = body_size
        self.crc = crc  #: may be null
        self.crc_matched = crc_matched
        #: `FitChunk` or `None` (depends on ``keep_raw_chunks`` option)
        self.chunk = chunk


class FitCRC:
    frame_type = FIT_FRAME_CRC

    __slots__ = ('crc', 'matched', 'chunk')

    def __init__(self, crc, matched, chunk):
        self.crc = crc
        self.matched = matched

        #: `FitChunk` or `None` (depends on ``keep_raw_chunks`` option)
        self.chunk = chunk


class FitDefinitionMessage:
    frame_type = FIT_FRAME_DEFMESG

    __slots__ = (
        # record header
        'is_developer_data',
        'local_mesg_num',
        'time_offset',

        # payload
        'mesg_type',
        'global_mesg_num',
        'endian',
        'field_defs',
        'dev_field_defs',

        'chunk')

    def __init__(
            self, is_developer_data, local_mesg_num, time_offset, mesg_type,
            global_mesg_num, endian, field_defs, dev_field_defs, chunk):
        self.is_developer_data = is_developer_data
        self.local_mesg_num = local_mesg_num
        self.time_offset = time_offset
        self.mesg_type = mesg_type
        self.global_mesg_num = global_mesg_num
        self.endian = endian
        self.field_defs = field_defs  #: list of `FieldDefinition`
        self.dev_field_defs = dev_field_defs  #: list of `DevFieldDefinition`
        #: `FitChunk` or `None` (depends on ``keep_raw_chunks`` option)
        self.chunk = chunk

    @property
    def name(self):
        if self.mesg_type is not None:
            return self.mesg_type.name
        else:
            return f'unknown_{self.global_mesg_num}'

    @property
    def all_field_defs(self):
        if not self.dev_field_defs:
            return self.field_defs
        return itertools.chain(self.field_defs, self.dev_field_defs)


class FitDataMessage:
    frame_type = FIT_FRAME_DATAMESG

    __slots__ = (
        # record header
        'is_developer_data',
        'local_mesg_num',
        'time_offset',

        'def_mesg',
        'fields',
        'chunk')

    def __init__(
            self, is_developer_data, local_mesg_num, time_offset, def_mesg,
            fields, chunk):
        #: Is this a "developer" message?
        self.is_developer_data = is_developer_data

        #: The **local** definition number of this message
        self.local_mesg_num = local_mesg_num

        #: Time offset in case header was compressed. `None` otherwise.
        self.time_offset = time_offset

        #: `FitDefinitionMessage`
        self.def_mesg = def_mesg

        #: list of `FieldData`
        self.fields = fields

        #: `FitChunk` or `None` (depends on ``keep_raw_chunks`` option)
        self.chunk = chunk

    def __iter__(self):
        """Iterate over the `FieldData` object in this mesage"""
        return iter(self.fields)

    @property
    def name(self):
        """Message name"""
        return self.def_mesg.name

    @property
    def global_mesg_num(self):
        """The **global** definition number of this message"""
        return self.def_mesg.global_mesg_num

    @property
    def mesg_type(self):
        """The `MessageType` object this message is associated to"""
        return self.def_mesg.mesg_type

    def has_field(self, field_name_or_num):
        """
        Is the desired field present in this message?

        *field_name_or_num* is the name of the field (`str`), or its definition
        number (`int`).

        .. seealso:: `get_field`, `get_fields`, `get_value`, `get_values`
        """
        for field in self.fields:
            if field.is_named(field_name_or_num):
                return True

        return False

    def get_field(self, field_name_or_num, idx=0):
        """
        Get the desired `FieldData` object.

        *field_name_or_num* is the name of the field (`str`), or its definition
        number (`int`).

        *idx* is the zero-based index of the specified field **among other
        fields with the same name/number**. I.e. not the index of the field in
        the list of fields of this message. That is, ``idx=0`` is the first
        *field_name_or_num* field found in this message.

        *idx* is useful in case a message contains multiple fields with the same
        *field_name_or_num*.

        .. seealso:: `get_fields`, `get_value`, `get_values`, `has_field`
        """
        current_idx = -1
        for field in self.fields:
            if field.is_named(field_name_or_num):
                current_idx += 1
                if current_idx == idx:
                    return field

        raise KeyError(
            f'field "{field_name_or_num}" (idx #{idx}) not found in '
            f'message "{self.name}"')

    def get_fields(self, field_name_or_num):
        """
        Like `get_field` but **yield** every `FieldData` object matching
        *field_name_or_num* fields in this message - i.e. generator.

        .. seealso:: `get_field`, `get_value`, `get_values`, `has_field`
        """
        for field in self.fields:
            if field.is_named(field_name_or_num):
                yield field

    def get_raw_value(
            self, field_name_or_num, *, idx=0, fallback=_UNSET, raw_value=True,
            fit_type=None, py_type=_UNSET):
        return self.get_value(
            field_name_or_num, idx=idx, fallback=fallback, raw_value=raw_value,
            fit_type=fit_type, py_type=py_type)

    def get_value(
            self, field_name_or_num, *, idx=0, fallback=_UNSET, raw_value=False,
            fit_type=None, py_type=_UNSET):
        """
        Get the value (or raw_value) of a field specified by its name or its
        definition number (*field_name_or_num*), with optional type checking.

        *idx* has the same meaning than for `get_field`.

        *fallback* can be specified to avoid `KeyError` being raised in case no
        field matched *field_name_or_num*.

        *fit_type* can be a `str` to indicate a given FIT type is expected (as
        defined in FIT profile; e.g. ``date_time``, ``manufacturer``, ...), in
        which case `TypeError` may be raised in case of a type mismatch.

        *py_type* can be a Python type or a `tuple` of types to expect (as
        passed to `isinstance`), in which case `TypeError` may be raised in case
        of a type mismatch.

        *raw_value* can be set to a true value so that the returned value is
        field's ``raw_value`` property instead of ``value``. This does not
        impact the way *fit_type* and *py_type* are interpreted.

        Special case: *field_name_or_num* can be `None`, in which case the field
        will be selected using *idx* only. In this case, *idx* is interpreted to
        be the zero-based index in the list of fields.

        .. seealso:: `get_values`, `get_field`, `get_fields`, `has_field`
        """
        assert (
            fit_type is _UNSET or
            fit_type is None or
            isinstance(fit_type, str))

        field_data = None

        if field_name_or_num is None:
            try:
                field_data = self.fields[idx]
                field_name_or_num = field_data.name_or_num

                # change the representation of idx so that its meaning can be
                # differentiated in case an exception is raised and it has to be
                # printed later on
                idx = f'[{idx}]'
            except KeyError:
                # KeyError exception is handled below so that the *fallback*
                # argument can be honored
                pass
        else:
            current_idx = -1
            for field in self.fields:
                if field.is_named(field_name_or_num):
                    current_idx += 1
                    if current_idx == idx:
                        field_data = field
                        break

        if field_data is None:
            if fallback is _UNSET:
                raise KeyError(
                    f'field "{field_name_or_num}" (idx #{idx}) not found in '
                    f'message "{self.name}"')
            return fallback

        # check FIT type if needed
        if fit_type and field_data.type.name != fit_type:
            raise TypeError(
                'unexpected type for FIT field '
                f'"{self.name}.{field_name_or_num}" (idx #{idx}; '
                f'got {field_data.type.name} instead of {fit_type})')

        # pick the right property
        value = field_data.value if not raw_value else field_data.raw_value

        # check value's type if needed
        if py_type is not _UNSET and not isinstance(value, py_type):
            if isinstance(py_type, (tuple, list)):
                py_type_str = ' or '.join([str(type(t)) for t in py_type])
            else:
                py_type_str = str(type(py_type))

            raise TypeError(
                'unexpected type for FIT value '
                f'"{self.name}.{field_name_or_num}" (idx #{idx}; '
                f'got {type(value)} instead of {py_type_str})')

        return value

    def get_values(
            self, field_name_or_num, *, raw_value=False, fit_type=None,
            py_type=_UNSET):
        """
        Like `get_value` but **yield** every value of all the fields that match
        *field_name_or_num* - i.e. generator.

        It is not possible to specify a *fallback* value so `KeyError` will
        always be raised in case the specified field was not found.

        The other arguments have the same meaning than for `get_value`.

        .. seealso:: `get_value`, `get_field`, `get_fields`, `has_field`
        """
        for idx, field_data in enumerate(self.fields):
            if field_data.is_named(field_name_or_num):
                value = self.get_value(
                    None, idx=idx, raw_value=raw_value,
                    fit_type=fit_type, py_type=py_type)
                yield value



# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT


class FitError(Exception):
    pass


class FitHeaderError(FitError):
    pass


class FitCRCError(FitError):
    pass


class FitEOFError(FitError):
    def __init__(self, expected, got, offset, message=''):
        self.expected = expected  #: number of expected bytes
        self.got = got  #: number of bytes read
        self.offset = offset  #: the file offset from which reading took place

        desc = f'expected {self.expected} bytes, got {self.got} @ {self.offset}'
        if not message:
            message = desc
        else:
            message += ' (' + desc + ')'

        super().__init__(message)


class FitParseError(FitError):
    def __init__(self, offset, message=''):
        self.offset = offset  #: the file offset from which reading took place

        desc = 'FIT parsing error @ ' + str(offset)
        if message:
            desc += ': ' + message

        super().__init__(desc)



# Copyright (c) Jean-Charles Lefebvre
# SPDX-License-Identifier: MIT

import enum
import io
import struct
import warnings

from . import processors
from . import profile
from . import records
from . import types
from . import utils
from .exceptions import FitCRCError, FitEOFError, FitHeaderError, FitParseError

__all__ = ['CrcCheck', 'ErrorHandling', 'FitReader']

_UNSET = object()


class CrcCheck(enum.Enum):
    """
    Defines the values expected by the ``check_crc`` parameter of `FitReader`'s
    constructor.
    """

    #: CRC is not computed at all (fastest).
    #: :class:`fitdecode.FitCRC` frame will still be yielded if present in the
    #: source FIT stream, but with meaningless values. In which case data
    #: processor's `fitdecode.DataProcessorBase.on_crc` method will still be
    #: called as well.
    DISABLED = 0

    #: CRC is computed but `FitReader` will never try to match CRCs. So no
    #: :class:`fitdecode.FitCRCError` will ever be raised.
    READONLY = 1

    #: CRC is computed and `FitReader` emits a warning message via
    #: `warnings.warn()` in case of mismatching CRC value. No
    #: :class:`fitdecode.FitCRCError` will ever be raised.
    WARN = 2

    #: CRC is computed and matched by `FitReader`.
    #: :class:`fitdecode.FitCRCError` is raised upon incorrect CRC value.
    RAISE = 3

    #: Alias of `RAISE` for backward compatibility.
    ENABLED = RAISE


class ErrorHandling(enum.Enum):
    """
    Defines the values expected by the ``error_handling`` parameter of
    `FitReader`'s constructor.
    """

    #: Disable developer types checking and other parsing error due to malformed
    #: FIT data. Parser proceeds when possible.
    #:
    #: For instance, if a `FitDefinitionMessage` references an undefined
    #: so-called *developer type* (or an undefined field), `FitReader` will
    #: silently ignore the error and will default to ``BYTE`` for the subsequent
    #: `FitDataMessage` frames that rely on this definition.
    IGNORE = 0

    #: Default behavior. Same as `IGNORE` but `FitReader` emits a warning with
    #: `warnings.warn()`.
    WARN = 1

    #: Strict mode. `FitReader` raises a `FitParseError` exception when
    #: malformed data is found, thus stopping the parsing of the current stream.
    RAISE = 2


class RecordHeader:
    __slots__ = (
        'is_definition', 'is_developer_data', 'local_mesg_num', 'time_offset')

    def __init__(self, is_definition, is_developer_data, local_mesg_num,
                 time_offset):
        self.is_definition = is_definition
        self.is_developer_data = is_developer_data
        self.local_mesg_num = local_mesg_num
        self.time_offset = time_offset


class FitReader:
    """
    Parse the content of a FIT stream or storage.

    Transparently supports "chained FIT Files" as per SDK's definition. A
    `FitHeader` object is yielded during iteration to mark the beginning of each
    new "FIT File".

    First argument *fileish* can be passed as a file-like or a path-like object
    (`os.PathLike`). File-like object must be opened in byte-mode. File-like
    object is not owned by `FitReader` so it is up to the caller to close it
    manually.

    Usage::

        import fitdecode

        with fitdecode.FitReader('file.fit') as fit:
            for frame in fit:
                # The yielded frame object is of one of the following types:
                # * fitdecode.FitHeader (FIT_FRAME_HEADER)
                # * fitdecode.FitDefinitionMessage (FIT_FRAME_DEFINITION)
                # * fitdecode.FitDataMessage (FIT_FRAME_DATA)
                # * fitdecode.FitCRC (FIT_FRAME_CRC)

                if frame.frame_type == fitdecode.FIT_FRAME_DATA:
                    # Here, frame is a FitDataMessage object.
                    # A FitDataMessage object contains decoded values that
                    # are directly usable in your script logic.
                    print(frame.name)

    Data processing:

    * You can specify your own data processor object using the *processor*
      argument.
    * The argument can be left untouched so that `DefaultDataProcessor` is used.
    * Otherwise, it can be set to `None` or any other false value to skip data
      processing entirely. This can speed up things a bit if your intent is only
      to manipulate the file at binary level (i.e. chunks), in which case
      *keep_raw_chunks* must be set to true.

    Raw chunks:

    * "raw chunk" or sometimes "frame", is the name given in fitdecode to the
      `bytes` block that represents one of the four FIT entities: `FitHeader`,
      `FitDefinitionMessage`, `FitDataMessage` and `FitCRC`.
    * While iterating a file with `FitReader`, you can for instance cut, stitch
      and/or reconstruct the file being read by using the
      `FitChunk` object attached to any of the four aforementioned entities, as
      long as the *keep_raw_chunks* option is true.

    Data bag:

    * A *data_bag* object can be passed to the constructor and then be retrieved
      via the `data_bag` property.
    * *data_bag* can be of any type (a `dict` by default) and will never be
      altered by this class.
    * A "data bag" is useful if you wish to store some context-sensitive data
      during the decoding of a file.
    * A typical use case is from a data processor that cannot hold its own
      context-sensitive data due to its instance being shared with other readers
      and/or by multiple threads (typically `DefaultDataProcessor`).

    """

    def __init__(
            self, fileish, *, processor=_UNSET, check_crc=CrcCheck.WARN,
            error_handling=ErrorHandling.WARN, keep_raw_chunks=False,
            data_bag=_UNSET):
        # backward compatibility
        if check_crc is True:
            check_crc = CrcCheck.RAISE
        elif check_crc is False:
            check_crc = CrcCheck.DISABLED

        assert isinstance(check_crc, CrcCheck)
        assert isinstance(error_handling, ErrorHandling)

        # modifiable options (public)
        self.check_crc = check_crc
        self.error_handling = error_handling

        # state (public)
        #: the *data_bag* object that was passed to the constructor, or, by
        #: default, a `dict` object
        self.data_bag = {} if data_bag is _UNSET else data_bag

        # immutable options (private)
        if processor is _UNSET:
            self._processor = processors.DefaultDataProcessor()
        else:
            self._processor = processor
        self._keep_raw = keep_raw_chunks

        # per-stream state (private)
        self._fd = None  # the file object to read from
        self._fd_owned = None  # do we own self._fd?
        self._read_offset = 0  # read cursor position in the file
        self._read_size = 0  # count bytes read from this file so far in total
        self._fit_file_index = -1  # the index of the current FIT file in this data stream  # noqa: E501

        # per-chunk state (private)
        self._chunk_index = 0   # the index number of the current chunk that is currently being read  # noqa: E501
        self._chunk_offset = 0  # the offset of the current chunk (relative to `read_offset`)  # noqa: E501
        self._chunk_size = 0    # the size of the current chunk

        # per-FIT-file state (private)
        self._crc = utils.CRC_START  # current CRC value, updated upon every read, reset on each new "FIT file"  # noqa: E501
        self._header = None  # `FitHeader` of the **current** "FIT file"
        self._current_file_id = None  # current file_id `FitDataMessage` object
        self._body_bytes_left = 0  # the number of bytes that are still to read before reaching the CRC footer of the current "FIT file"  # noqa: E501
        self._local_mesg_defs = {}  # registry of every `FitDefinitionMessage` in this file so far  # noqa: E501
        self._local_dev_types = {}  # registry of developer types
        self._compressed_ts_accumulator = 0  # state value for the so-called "Compressed Timestamp Header"  # noqa: E501
        self._accumulators = {}
        self._last_timestamp = 0
        self._hr_start_timestamp = 0  # special case for the ``hr`` message

        if hasattr(fileish, 'read'):
            self._fd = fileish
            self._fd_owned = False
        elif isinstance(fileish, str) or hasattr(fileish, '__fspath__'):
            self._fd = open(fileish, mode='rb')
            self._fd_owned = True
        else:
            self._fd = io.BytesIO(fileish)
            self._fd_owned = True

        try:
            self._read_offset = self._fd.tell()
            self._chunk_offset = self._read_offset
        except (AttributeError, OSError):
            pass

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        return self.close()

    def __iter__(self):
        yield from self._read_next()

    @property
    def processor(self):
        """Read-only access to the data processor object."""
        return self._processor

    @property
    def last_header(self):
        """The last read `FitHeader` object. May be `None`."""
        return self._header

    @property
    def last_timestamp(self):
        """
        The last ``timestamp`` value (`int` or `float`).

        Often useful in FIT files since some data fields rely on it like
        ``timestamp_16`` and ``timestamp_ms`` for instance.

        Hint: you usually want to use this property from your own processor
        class derived from on of the processors available from
        `fitdecode.processors`.
        """
        return self._last_timestamp

    @property
    def file_id(self):
        """
        The ``file_id`` `FitDataMessage` object related to the current FIT file.

        May be `None`. Typically before a `FitHeader` frame, or after a `FitCRC`
        frame.
        """
        return self._current_file_id

    @property
    def fit_file_index(self):
        """
        The zero-based index `int` of the so-called *FIT file* currently read
        from this data stream so far.

        `None` if no *FIT file* header has been encoutered yet.
        """
        return self._fit_file_index if self._fit_file_index >= 0 else None

    @property
    def fit_files_count(self):
        """The number of FIT files found in this data stream so far."""
        return self._fit_file_index + 1

    @property
    def local_mesg_defs(self):
        """
        Read-only access to the `dict` of local message types of the current
        "FIT file".

        It is cleared by `close()` (or ``__exit__()``), and also each time a FIT
        file header is reached (i.e. at the beginning of a file, or after a
        `FitCRC`).
        """
        return self._local_mesg_defs

    @property
    def local_dev_types(self):
        """
        Read-only access to the `dict` of developer types of the current
        "FIT file".

        It is cleared by `close()` (or ``__exit__()``), and also each time a FIT
        file header is reached (i.e. at the beginning of a file, or after a
        `FitCRC`).
        """
        return self._local_dev_types

    def close(self):
        """
        Close the internal file handle if it is owned by this object, and clear
        the internal state.
        """
        if self._fd is not None and self._fd_owned and hasattr(self._fd, 'close'):
            self._fd.close()

        self._fd = None
        self._fd_owned = None
        self._read_offset = 0
        self._read_size = 0
        self._fit_file_index = -1
        self._chunk_index = 0
        self._chunk_offset = 0
        self._chunk_size = 0
        self._crc = utils.CRC_START
        self._header = None
        self._current_file_id = None
        self._body_bytes_left = 0
        self._local_mesg_defs = {}
        self._local_dev_types = {}
        self._compressed_ts_accumulator = 0
        self._accumulators = {}
        self._last_timestamp = 0
        self._hr_start_timestamp = 0

    # ONLY PRIVATE METHODS BELOW ***********************************************

    def _read_next(self):

        def _update_state():
            if self._fd is not None:
                self._chunk_index += 1
                self._chunk_offset += self._chunk_size
                self._chunk_size = 0

        while self._fd is not None:
            assert self._chunk_size == 0

            if self._header is None:
                assert self._body_bytes_left == 0

                self._reset_per_fit_state()
                self._read_header()
                if self._header is None:
                    break

                self._fit_file_index += 1

                yield self._header
                _update_state()

            elif self._body_bytes_left > 0:
                assert self._header is not None

                record = self._read_record()

                assert isinstance(record, (
                    records.FitDefinitionMessage,
                    records.FitDataMessage))

                assert self._chunk_size <= self._body_bytes_left
                self._body_bytes_left -= self._chunk_size

                yield record
                _update_state()

            else:
                assert self._header is not None
                assert self._body_bytes_left == 0

                try:
                    crc_obj = self._read_crc()
                except FitEOFError:
                    # if self.check_crc is not CrcCheck.RAISE:
                    #     # There is no CRC footer in this file (or it is
                    #     # incomplete) but caller does not mind about CRC so
                    #     # we will just ignore this
                    #     break
                    raise

                yield crc_obj
                _update_state()

                # We've reached the end of this FIT file... To avoid incorrect
                # behavior due to malformed FIT stream (i.e. next FIT header
                # missing), reset the internal state now as well, instead of
                # resetting it only when a FIT header is read.
                self._reset_per_fit_state()

    def _reset_per_fit_state(self):
        # reset per-FIT-file state
        self._crc = utils.CRC_START
        self._header = None
        self._current_file_id = None
        self._body_bytes_left = 0
        self._local_mesg_defs = {}
        self._local_dev_types = {}
        self._compressed_ts_accumulator = 0
        self._accumulators = {}
        self._last_timestamp = 0
        self._hr_start_timestamp = 0

    def _read_header(self):
        try:
            chunk, header_size, proto_ver, profile_ver, body_size, \
                header_magic = self._read_struct('<2BHI4s')
        except FitEOFError as exc:
            if not exc.got:
                # regular EOF: storage is empty or previous "FIT file" ended
                # normally
                return
            raise FitHeaderError(f'file truncated? {exc}')

        # check header size
        if header_size < len(chunk) or header_magic != b'.FIT':
            raise FitHeaderError(f'not a FIT file @ {self._chunk_offset}')

        # read the extended part of the header (i.e. byte 12-...)
        extra_header_size = header_size - len(chunk)
        read_crc = None
        crc_matched = None
        if extra_header_size:
            # at least 2 bytes expected for the CRC
            if extra_header_size < 2:
                raise FitHeaderError(
                    f'unsupported FIT header (CRC field missing; '
                    f'header offset: {self._chunk_offset})')

            extra_chunk = self._read_bytes(extra_header_size)
            if len(extra_chunk) != extra_header_size:
                raise FitHeaderError(
                    f'truncated FIT header '
                    f'(header offset: {self._chunk_offset})')

            (read_crc, ) = struct.unpack('<H', extra_chunk)
            if not read_crc:  # can be null according to SDK
                read_crc = None
                crc_matched = None
            else:
                computed_crc = utils.compute_crc(chunk)
                crc_matched = computed_crc == read_crc

                if not crc_matched and (
                        self.check_crc is CrcCheck.WARN or
                        self.check_crc is CrcCheck.RAISE):
                    msg = (
                        f'mismatching CRC in FIT header '
                        f'(header offset: {self._chunk_offset})')

                    if self.check_crc is CrcCheck.RAISE:
                        raise FitCRCError(msg)
                    elif self.check_crc is CrcCheck.WARN:
                        warnings.warn(msg)

            chunk += extra_chunk

        proto_ver = (proto_ver >> 4, proto_ver & ((1 << 4) - 1))
        profile_ver = (int(profile_ver / 100), int(profile_ver % 100))

        # update state
        self._header = records.FitHeader(
            header_size=header_size,
            proto_ver=proto_ver,
            profile_ver=profile_ver,
            body_size=body_size,
            crc=read_crc,
            crc_matched=crc_matched,
            chunk=self._keep_chunk(chunk))
        self._body_bytes_left = body_size

        if self._processor:
            self._processor.on_header(self, self._header)

    def _read_crc(self):
        computed_crc = self._crc
        chunk, read_crc = self._read_struct('<H')

        if computed_crc != read_crc and (
                self.check_crc is CrcCheck.WARN or
                self.check_crc is CrcCheck.RAISE):
            msg = (
                f'mismatching CRC in FIT footer '
                f'(footer offset: {self._chunk_offset}; '
                f'read crc: {read_crc}; '
                f'expected crc: {computed_crc})')

            if self.check_crc is CrcCheck.RAISE:
                raise FitCRCError(msg)
            elif self.check_crc is CrcCheck.WARN:
                warnings.warn(msg)

        crc_obj = records.FitCRC(
            read_crc,
            computed_crc == read_crc,
            self._keep_chunk(chunk))

        if self._processor:
            self._processor.on_crc(self, crc_obj)

        return crc_obj

    def _read_record(self):
        # read header
        chunk = self._read_bytes(1)
        if chunk[0] & 0x80:  # bit 7: compressed timestamp?
            record_header = RecordHeader(
                False,                  # is_definition
                False,                  # is_developer_data
                (chunk[0] >> 5) & 0x3,  # local_mesg_num; bits 5-6
                chunk[0] & 0x1f)        # time_offset; bits 0-4
        else:
            record_header = RecordHeader(
                bool(chunk[0] & 0x40),  # is_definition; bit 6
                bool(chunk[0] & 0x20),  # is_developer_data; bit 5
                chunk[0] & 0xf,         # local_mesg_num; bits 0-3
                None)                   # time_offset

        # read record payload
        if record_header.is_definition:
            message = self._read_definition_message(chunk, record_header)
        else:
            message = self._read_data_message(chunk, record_header)

            if message.mesg_type is not None:
                if message.mesg_type.mesg_num == profile.MESG_NUM_DEVELOPER_DATA_ID:
                    self._add_dev_data_id(message)
                elif message.mesg_type.mesg_num == profile.MESG_NUM_FIELD_DESCRIPTION:
                    self._add_dev_field_description(message)

        return message

    def _read_definition_message(self, header_chunk, record_header):
        record_chunks = [header_chunk]

        # read the "fixed content" part
        extra_chunk = self._read_bytes(5)
        record_chunks.append(extra_chunk)
        endian = '<' if not extra_chunk[1] else '>'
        global_mesg_num, num_fields = struct.unpack(
            f'{endian}2xHB', extra_chunk)

        # get global message's declaration from our profile if any
        mesg_type = profile.MESSAGE_TYPES.get(global_mesg_num)

        field_unpacker = struct.Struct(f'{endian}3B')
        field_defs = []
        dev_field_defs = []

        # read field definitions
        for _ in range(num_fields):
            extra_chunk = self._read_bytes(field_unpacker.size)
            record_chunks.append(extra_chunk)

            field_def_num, field_size, base_type_num = \
                field_unpacker.unpack(extra_chunk)

            field = mesg_type.fields.get(field_def_num) if mesg_type else None
            base_type = types.BASE_TYPES.get(
                base_type_num, types.BASE_TYPE_BYTE)

            if (field_size % base_type.size) != 0:
                # should we fall back to byte encoding instead raising a
                # FitParseError?
                # well, apparently yes:
                #   https://github.com/polyvertex/fitdecode/issues/13
                #   https://github.com/dtcooper/python-fitparse/pull/116
                #   https://github.com/GoldenCheetah/GoldenCheetah/issues/3645
                msg = (
                    f'invalid field size {field_size} in definition message @ '
                    f'{self._chunk_offset} for type {base_type.name} (expected '
                    f'a multiple of {base_type.size})')

                if self.error_handling is ErrorHandling.RAISE:
                    raise FitParseError(self._chunk_offset, msg)
                elif self.error_handling is ErrorHandling.WARN:
                    warnings.warn(msg)
                else:
                    assert self.error_handling is ErrorHandling.IGNORE

                base_type = types.BASE_TYPE_BYTE

            # if the field has components that are accumulators,
            # start recording their accumulation at 0
            if field and field.components:
                for component in field.components:
                    if component.accumulate:
                        accumulators = \
                            self._accumulators.setdefault(global_mesg_num, {})
                        accumulators[component.def_num] = 0

            field_defs.append(types.FieldDefinition(
                field, field_def_num, base_type, field_size))

        # read developer field definitions if any
        if record_header.is_developer_data:
            # read the number of developer fields definitions that follow
            extra_chunk = self._read_bytes(1)
            record_chunks.append(extra_chunk)
            num_dev_fields = extra_chunk[0]

            # read field definitions
            for _ in range(num_dev_fields):
                extra_chunk = self._read_bytes(field_unpacker.size)
                record_chunks.append(extra_chunk)

                field_def_num, field_size, dev_data_index = \
                    field_unpacker.unpack(extra_chunk)

                field = self._get_dev_type(
                    record_header.local_mesg_num, global_mesg_num,
                    dev_data_index, field_def_num)

                dev_field_defs.append(types.DevFieldDefinition(
                    field, dev_data_index, field_def_num, field_size))

        def_mesg = records.FitDefinitionMessage(
            record_header.is_developer_data,
            record_header.local_mesg_num,
            record_header.time_offset,
            mesg_type,
            global_mesg_num,
            endian,
            field_defs,
            dev_field_defs,
            self._keep_chunk(record_chunks))

        # According to FIT protocol's specification (section 4.8.3), it is ok to
        # redefine message types
        self._local_mesg_defs[record_header.local_mesg_num] = def_mesg

        return def_mesg

    def _read_data_message(self, header_chunk, record_header):
        record_chunks = [header_chunk]

        try:
            def_mesg = self._local_mesg_defs[record_header.local_mesg_num]
        except KeyError:
            raise FitParseError(
                self._chunk_offset,
                f'local message {record_header.local_mesg_num} not defined')

        extra_chunks, raw_values = self._read_data_message_raw_values(def_mesg)
        record_chunks.extend(extra_chunks)
        message_fields = []

        for field_def, raw_value in zip(def_mesg.all_field_defs, raw_values):

            field, parent_field = field_def.field, None
            if field:
                field, parent_field = self._resolve_subfield(
                    field, def_mesg, raw_values)

                # resolve component fields
                if field.components:
                    # special case for hr.event_timestamp_12
                    is_hr_event_timestamp_12 = (
                        def_mesg.global_mesg_num == profile.MESG_NUM_HR and
                        not field_def.is_dev and
                        field_def.def_num == profile.FIELD_NUM_HR_EVENT_TIMESTAMP_12)

                    for component in field.components:
                        # render its raw value
                        try:
                            cmp_raw_value = component.render(raw_value)
                        except ValueError:
                            continue

                        # apply accumulated value
                        if component.accumulate and cmp_raw_value is not None:
                            accumulator = self._accumulators[
                                def_mesg.global_mesg_num]

                            cmp_raw_value = self._apply_compressed_accumulation(
                                cmp_raw_value,
                                accumulator[component.def_num],
                                component.bits)

                            accumulator[component.def_num] = cmp_raw_value

                        # apply scale and offset from component, not from the
                        # dynamic field as they may differ
                        cmp_raw_value = self._apply_scale_offset(
                            component, cmp_raw_value)

                        # extract the component's dynamic field from def_mesg
                        cmp_field = def_mesg.mesg_type.fields[component.def_num]

                        # resolve a possible subfield
                        cmp_field, cmp_parent_field = self._resolve_subfield(
                            cmp_field, def_mesg, raw_values)
                        cmp_value = cmp_field.render(cmp_raw_value)

                        # special case: hr.event_timestamp_12
                        if is_hr_event_timestamp_12:
                            assert self._hr_start_timestamp > 0
                            cmp_value += self._hr_start_timestamp

                        message_fields.append(types.FieldData(
                            None,              # field_def
                            cmp_field,         # field
                            cmp_parent_field,  # parent_field
                            cmp_value,         # value
                            cmp_raw_value))    # raw_value

                decoded_value = self._apply_scale_offset(
                    field, field.render(raw_value))
            else:
                decoded_value = raw_value

            # specifics
            if (field_def.def_num == profile.FIELD_NUM_TIMESTAMP and
                    raw_value is not None):
                self._last_timestamp = decoded_value
                # update compressed timestamp field
                self._compressed_ts_accumulator = raw_value
            elif (def_mesg.global_mesg_num == profile.MESG_NUM_HR and
                    not field_def.is_dev and
                    field_def.def_num == profile.FIELD_NUM_HR_EVENT_TIMESTAMP):
                # hr.event_timestamp_12 fields are accumulated from an initial
                # hr.event_timestamp value
                # assert self._last_timestamp > 0
                self._hr_start_timestamp = self._last_timestamp

            message_fields.append(types.FieldData(
                field_def,      # field_def
                field,          # field
                parent_field,   # parent_field
                decoded_value,  # value
                raw_value))     # raw_value

        # apply timestamp field if we got a header
        if record_header.time_offset is not None:
            ts_value = self._apply_compressed_accumulation(
                record_header.time_offset, self._compressed_ts_accumulator, 5)

            self._compressed_ts_accumulator = ts_value

            message_fields.append(types.FieldData(
                None,                                           # field_def
                profile.FIELD_TYPE_TIMESTAMP,                   # field
                None,                                           # parent_field
                profile.FIELD_TYPE_TIMESTAMP.render(ts_value),  # value
                ts_value))                                      # raw_value

        # apply data processors
        if self._processor:
            for field_data in message_fields:
                self._processor.on_process_type(self, field_data)
                self._processor.on_process_field(self, field_data)
                self._processor.on_process_unit(self, field_data)

        data_message = records.FitDataMessage(
            record_header.is_developer_data,
            record_header.local_mesg_num,
            record_header.time_offset,
            def_mesg,
            message_fields,
            self._keep_chunk(record_chunks))

        if self._processor:
            self._processor.on_process_message(self, data_message)

        # keep track of the last file_id message
        if def_mesg.global_mesg_num == profile.MESG_NUM_FILE_ID:
            self._current_file_id = data_message

        return data_message

    def _read_data_message_raw_values(self, def_mesg):
        raw_values = []
        record_chunks = []

        for field_def in def_mesg.all_field_defs:
            base_type = field_def.base_type

            # struct format to read "[N]" base types
            unpacker = struct.Struct(
                f'{def_mesg.endian}'
                f'{int(field_def.size / base_type.size)}'
                f'{base_type.fmt}')

            # read the chunk
            chunk = self._read_bytes(unpacker.size)
            record_chunks.append(chunk)

            # extract the raw value
            raw_value = unpacker.unpack(chunk)

            if base_type.identifier == types.BASE_TYPE_BYTE.identifier:
                raw_value = base_type.parse(raw_value)
            elif len(raw_value) > 1:
                # If the field returns with a tuple of values it's definitely an
                # oddball, but we'll parse it on a per-value basis. If it's a
                # byte type, treat the tuple as a single value
                raw_value = tuple(base_type.parse(v) for v in raw_value)
            else:
                # Otherwise, just scrub the singular value
                raw_value = base_type.parse(raw_value[0])

            raw_values.append(raw_value)

        return record_chunks, raw_values

    def _read_struct(self, fmt, *, endian=None):
        assert fmt
        if endian:
            fmt = endian + fmt

        unpacker = struct.Struct(fmt)
        if unpacker.size <= 0:
            raise ValueError(f'invalid struct format "{fmt}"')

        chunk = self._read_bytes(unpacker.size)

        return (chunk, ) + unpacker.unpack(chunk)

    def _read_bytes(self, size):
        if size <= 0:
            raise ValueError('size')

        chunk = utils.blocking_read(self._fd, size)
        chunk_size = 0 if not chunk else len(chunk)
        if chunk_size != size:
            raise FitEOFError(size, chunk_size, self._read_offset)

        if chunk:
            if self.check_crc is not CrcCheck.DISABLED:
                self._crc = utils.compute_crc(chunk, crc=self._crc)
            self._chunk_size += chunk_size
            self._read_offset += chunk_size
            self._read_size += chunk_size

        return chunk

    def _keep_chunk(self, chunk):
        if not self._keep_raw:
            return None

        assert chunk

        if isinstance(chunk, (list, tuple)):
            # *chunk* is a iterable of chunks
            chunk = b''.join(chunk)
        else:
            assert isinstance(chunk, bytes)

        assert len(chunk) == self._chunk_size

        return records.FitChunk(
            self._chunk_index, self._chunk_offset, chunk)

    def _add_dev_data_id(self, message):
        try:
            dev_data_index = message.get_raw_value('developer_data_index')
            dev_data_index = int(dev_data_index)
        except KeyError as exc:
            msg = (
                f'{exc} (local_mesg_num: {message.local_mesg_num}; '
                f'chunk_offset: {self._chunk_offset})')

            if self.error_handling is ErrorHandling.RAISE:
                raise FitParseError(self._chunk_offset, msg)
            elif self.error_handling is ErrorHandling.WARN:
                msg += '; adding dummy dev data...'
                warnings.warn(msg)
            else:
                assert self.error_handling is ErrorHandling.IGNORE

            dev_data_index = None

        application_id = message.get_raw_value('application_id', fallback=None)

        # declare/overwrite type
        self._add_dev_data_id_impl(dev_data_index, application_id)

    def _add_dev_data_id_impl(self, dev_data_index, application_id=None):
        self._local_dev_types[dev_data_index] = {
            'dev_data_index': dev_data_index,
            'application_id': application_id,
            'fields': {}}

    def _add_dev_field_description(self, message):
        adfdi_args = []

        # CAUTION: order of args matters, check out below how adfdi_args is used
        for raw_value_name in (
                'developer_data_index', 'field_definition_number',
                'fit_base_type_id', 'field_name', 'units', 'native_field_num'):
            try:
                raw_value = message.get_raw_value(raw_value_name)
            except KeyError as exc:
                # assume some fields are optional (issues #22 and #24 at least)
                if raw_value_name in (
                        'developer_data_index', 'field_definition_number'):
                    msg = (
                        f'{exc} (local_mesg_num: {message.local_mesg_num}; '
                        f'chunk_offset: {self._chunk_offset})')

                    if self.error_handling is ErrorHandling.RAISE:
                        raise FitParseError(self._chunk_offset, msg)
                    elif self.error_handling is ErrorHandling.WARN:
                        msg += '; adding dummy dev data...'
                        warnings.warn(msg)
                    else:
                        assert self.error_handling is ErrorHandling.IGNORE

                raw_value = None

            if raw_value_name == 'developer_data_index':
                if raw_value is not None:
                    raw_value = int(raw_value)
                if raw_value not in self._local_dev_types:
                    raise FitParseError(
                        self._chunk_offset,
                        f'developer_data_index {raw_value} not defined')
            elif raw_value_name == 'fit_base_type_id':
                if raw_value is None:
                    raw_value = types.BASE_TYPE_BYTE
                else:
                    raw_value = types.BASE_TYPES[raw_value]

            adfdi_args.append(raw_value)

        # declare/overwrite type
        self._add_dev_field_description_impl(*adfdi_args)

    def _add_dev_field_description_impl(
            self, dev_data_index, field_def_num, base_type=types.BASE_TYPE_BYTE,
            name=None, units=None, native_field_num=None):
        self._local_dev_types[dev_data_index]['fields'][field_def_num] = \
            types.DevField(
                dev_data_index, name, field_def_num, base_type, units,
                native_field_num)

    def _get_dev_type(
            self, local_mesg_num, global_mesg_num, dev_data_index,
            field_def_num):
        try:
            dev_type = self._local_dev_types[dev_data_index]
        except KeyError:
            msg = (
                f'dev_data_index {dev_data_index} not defined (looking up for '
                f'field {field_def_num}; local_mesg_num: {local_mesg_num}; '
                f'global_mesg_num: {global_mesg_num}; chunk_offset: '
                f'{self._chunk_offset})')

            if self.error_handling is ErrorHandling.RAISE:
                raise FitParseError(self._chunk_offset, msg)
            elif self.error_handling is ErrorHandling.WARN:
                msg += '; adding dummy type...'
                warnings.warn(msg)
            else:
                assert self.error_handling is ErrorHandling.IGNORE

            self._add_dev_data_id_impl(dev_data_index)
            dev_type = self._local_dev_types[dev_data_index]

        try:
            return dev_type['fields'][field_def_num]
        except KeyError:
            msg = (
                f'no such field {field_def_num} for dev_data_index '
                f'{dev_data_index} (local_mesg_num: {local_mesg_num}; '
                f'global_mesg_num: {global_mesg_num}; chunk_offset: '
                f'{self._chunk_offset})')

            if self.error_handling is ErrorHandling.RAISE:
                raise FitParseError(self._chunk_offset, msg)
            elif self.error_handling is ErrorHandling.WARN:
                msg += '; defaulting to BYTE field type...'
                warnings.warn(msg)
            else:
                assert self.error_handling is ErrorHandling.IGNORE

            self._add_dev_field_description_impl(
                dev_data_index, field_def_num)

            return self._local_dev_types[dev_data_index]['fields'][field_def_num]

    @staticmethod
    def _resolve_subfield(field, def_mesg, raw_values):
        # resolve into (field, parent) ie (subfield, field) or (field, none)
        if field.subfields:
            for sub_field in field.subfields:
                # go through reference fields for this sub field
                for ref_field in sub_field.ref_fields:
                    # go through field defs AND their raw values
                    for field_def, raw_value in zip(
                            def_mesg.field_defs, raw_values):
                        # if there's a definition number AND raw value match on
                        # the reference field, then we return this subfield
                        if (field_def.def_num == ref_field.def_num and
                                ref_field.raw_value == raw_value):
                            return sub_field, field

        return field, None

    @staticmethod
    def _apply_compressed_accumulation(raw_value, accumulation, num_bits):
        max_value = 1 << num_bits
        max_mask = max_value - 1
        base_value = raw_value + (accumulation & ~max_mask)

        if raw_value < (accumulation & max_mask):
            base_value += max_value

        return base_value

    @classmethod
    def _apply_scale_offset(cls, field, raw_value):
        # apply numeric transformations (scale + offset)
        if isinstance(raw_value, tuple):
            # contains multiple values, apply transformations to all of them
            return tuple(cls._apply_scale_offset(field, x) for x in raw_value)
        elif isinstance(raw_value, (int, float)):
            if field.scale:
                raw_value = float(raw_value) / field.scale
            if field.offset:
                raw_value = raw_value - field.offset

        return raw_value
